/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: invPhi.cc,v 1.7 1992/11/30 17:31:31 bengt Exp $ */

/* 
 * Computes the inverse of Phi,  the normal distribution
 *
 * Reference: Kinder/Osius/Timm, Statistik fr Biologen und Mediziner
 * Coding: Bengt Mrtensson
 */

#include "externals.h"
#include "gnanslib.h"
#include <math.h>

inline static double
inv_phi_aux(double y)
{
  static double a0 = 0.322232431088;
  static double a1 = 1.0;
  static double a2 = 0.342242088547;
  static double a3 = 0.0204231210245;
  static double a4 = 0.453642210148e-4;
  static double b0 = 0.0993484626060;
  static double b1 = 0.588581570495;
  static double b2 = 0.531103462366;
  static double b3 = 0.103537752850;
  static double b4 = 0.0038568700634;

  register double t = sqrt(-2*log(y));

  return t - (a0 + a1*t + a2*t*t + a3*t*t*t + a4*t*t*t*t)/
             (b0 + b1*t + b2*t*t + b3*t*t*t + b4*t*t*t*t);
}

double
invPhi(double y)
{
  /* Don't check validity of input, for maximal speed */
  return y < 0.5 ? -inv_phi_aux(y) : inv_phi_aux(1-y);
}
