/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: normalizer.cc,v 1.2 1994/01/12 22:57:01 bengt Exp $ */

// This rounds its first argument to second-argument significant digits.
// If optional third argument is true, truncates instead of rounds.

// Not written to be particularly efficient...

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "gnanslib.h"

normalizer::normalizer()
{
  r = 1;
  max = invmax = 1;
};

void
normalizer::setlimit(double lim)
{
  max = lim;
  invmax = 1/max;
  r = 1;
};

double 
normalizer::previous_normalizations()
{
  return r;
};

void
normalizer::normalize(double &x)
{
  if (abs(x) > max || abs(x) < invmax) {
    r *= abs(x);
    x /= abs(x);
  }
};

void
normalizer::normalize(double &x1, double &x2)
{
  double rho = sqrt(sqr(x1) + sqr(x2));
  if (rho > max || rho < invmax) {
    r *= rho;
    x1 /= rho;
    x2 /= rho;
  }
};

void
normalizer::normalize(double &x1, double &x2, double &x3)
{
  double rho = sqrt(sqr(x1) + sqr(x2) + sqr(x3));
  if (rho > max || rho < invmax) {
    r *= rho;
    x1 /= rho;
    x2 /= rho;
    x3 /= rho;
  }
};

void
normalizer::normalize(double &x1, double &x2, double &x3, double &x4)
{
  double rho = sqrt(sqr(x1) + sqr(x2) + sqr(x3) + sqr(x4));
  if (rho > max || rho < invmax) {
    r *= rho;
    x1 /= rho;
    x2 /= rho;
    x3 /= rho;
    x4 /= rho;
  }
};

void
normalizer::normalize(double &x1, double &x2, double &x3, double &x4,
		      double &x5)
{
  double rho = sqrt(sqr(x1) + sqr(x2) + sqr(x3) + sqr(x4) + sqr(x5));
  if (rho > max || rho < invmax) {
    r *= rho;
    x1 /= rho;
    x2 /= rho;
    x3 /= rho;
    x4 /= rho;
    x5 /= rho;
  }
};

