/* -*- c++ -*-
Copyright  1992, 1993 Bengt Mrtensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: CIntegrator.h,v 1.4 1994/08/19 18:21:46 bengt Exp $

#ifndef _CIntegrator_h
#define _CIntegrator_h

#if __GNUC__
#pragma interface
#endif

#include "Integrator.h"
#include <string>

// For the function prototypes for external Fortran routines.
// Function defining the DE to be integrated
typedef void (*fp)(double *t, double *y, double *dy);

// Ditto, with four arguments
typedef void (*hairy_fp)(const int *n, double *t, double *y, double *dy);

// Data structure for integration error messages 
class integration_error_message {
  // Let these be public to allow initializers
  //private:
public:
  // integer, as returned from an integration algorithm 
  signed int errorno;
  
  // Human understandable (hopefully!) error message 
  const char * error_message;
  
public:
  signed int get_errorno() const { return errorno; };
  const char * get_error_message() const { return error_message; }
  
};

class ContinuousIntegrator : public Integrator {
public:
  // None
protected:

  // Check continuous time
  virtual bool continuous_time_barf();

  // Check that the step is nonzero
  bool barf_if_step_zero();

  // Check that the system is deterministic
  virtual bool barf_on_wieners();

  // Present an integration error
  void integration_error(double t1, double t2, int errorcode,
			 const integration_error_message *errors,
			 const char *msg = 0) const;

  // Have we computed beyond the final time?
  bool overshot() const;

  // Finished with the task (or interrupted by the user)
  virtual bool finished() const;

  // The function carrying out the main task
  // The version here is a generic driver for a dumb (fixed stepsize and no
  // error control) algorithm.
  virtual bool loop(bool initialize);

  // Wrapper around integrator call
  virtual bool call_integrator() = 0;

  // Is the task completed, free stepsize version
  bool more_to_do_zero_step() const;

  // Function defining the ODE, four argument version
  static void hairy_f(const int *n, double *t, double *x, double *dx);

  // Algorithm specific initializatins
  virtual void initialize_integration();

  // For integrators using the xerrwv call, this function makes the 
  // error messages visible.
  std::string xerrwv_flush();

public:
  static bool xerrwv_initialized;
  static std::string xerrwv_themessage;
};


#endif
