/* -*- c++ -*-
Copyright  1992, 1993 Bengt Mrtensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: DIntegrator.cc,v 1.7 1994/08/24 09:58:06 bengt Exp $

#ifdef __GNUC__
#pragma implementation
#endif

#include "externals.h"
#include "DIntegrator.h"

static bool
discrete_time_barf()
{
  if (current_system.system_type != discrete) {
    // Programming error
    gnans_error(impossible,
                "The selected algorithm only handles discrete time");
    return true;
  }
  if (starttime >= stoptime) {
    gnans_error(non_fatal,
                "stoptime must be larger than starttime");
    return true;
  }
  return false;
}

const char *
DiscreteIntegrator::documentation() const
{
  return "`Integration routine' for discrete time systems.";
}

const char *
DiscreteIntegrator::name() const
{
  return "DiscreteIntegrator";
}

void
DiscreteIntegrator::store_stuff()
{
  // Since this may be the first point after an alarm, check that we are
  // not saving the same point twice.

  if (last_t_stored != systemtime && ! floating_point_error_occured) {
    last_t_stored = systemtime;
    (*current_system.f_store_ptr)();
    if (sample_no >= trajectories_current_size)
      resize_trajectories();
    
    for (register unsigned int i = 0; i < number_of_stores; i++)
      stores[i].trajectory[sample_no] = *stores[i].ptr;
    sample_no++;
  }
}

bool 
DiscreteIntegrator::finished() const
{
  return stoptime - systemtime < 1 || ! running_integration;
}

bool
DiscreteIntegrator::more_to_do()
{
  return stoptime - systemtime >= 1
    && running_integration && ! has_been_alarmed;
}

bool
DiscreteIntegrator::initialize_integrator()
{
  Vect tempstates(current_system.number_of_states);
  
  setup_times();
  if (discrete_time_barf())
    return false;
  systemtime = double(starttime);
  
  tempstates = current_system.state_vector;
  (*current_system.discrete_f_ptr)();
  (*current_system.f_store_ptr)();
  (*current_system.f_accepted_ptr)();
  current_system.state_vector = tempstates;
  return true;
}

bool 
DiscreteIntegrator::loop(bool initialize)
{
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }

  while (more_to_do()) {
    store_if_necessary();
    (*current_system.discrete_f_ptr)();	// f updates y
    (*current_system.f_accepted_ptr)();
    ++systemtime;
  }
  
  if (finished())
    store_stuff();
  return finished();
}
