/*                -*- c++ -*-
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Integrator.h,v 1.5 1994/08/24 09:58:06 bengt Exp $

#ifndef _Integrator_h
#define _Integrator_h

#ifdef __GNUC__
#pragma interface
//**//#include <builtin.h>
#endif

#include "Enums.h"
#include "Vect.h"

class Integrator {
public:

  // Returns the name of the algorithm (as presented to the user)
  virtual const char *name() const = 0;
  operator const char *() const         { return name(); };

  // Function carrying out the integration 
  // The function takes a boolean argument, indicating whether initialization
  // should be done or not. This argument should be true on the first call,
  // false on subsequent. Scratch variables must be static if they are to
  // retain their values between subsequent calls.

  //  Returns true if the integration is completely finished, false otherwise.
  virtual bool loop(bool initialize) = 0;

  // Returns the documentation
  virtual const char *documentation() const = 0;

  // Doing Wiener processes?
  virtual bool stochastic() const     { return false; }

  virtual signed int default_method_flag() const { return 0; }
protected:
  double abs(double x) { return x > 0 ? x : -x; }
  double last_t_stored;
  double signed_step;

  Vect junk_vector;		// Scratch space

  bool more_to_do_nonzero_step() const;
  virtual void store_stuff();
  virtual bool finished() const;
  void setup_times();
  void store_if_necessary();
  virtual bool initialize_integrator() = 0;
};

#endif
