/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Lsode.cc,v 1.7 1994/11/01 12:51:48 bengt Exp bengt $

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "utillib.h"
#include "Lsode.h"

#ifdef FORTRAN_UNDERSCORES
#define lsode lsode_
#endif

#ifdef FORTRAN_UPPERCASE
#define lsode LSODE
#endif

// External declaration for the integrator
extern "C" void 
#ifdef __GNUG__
lsode(hairy_fp f, int *neq, double *y, double *t,
      double *tout, int *itol, double *rtol,
      double *atol, int *itask, int *istate,
      int *iopt, double *rwork, int *lrw,
      int *iwork, int *liw, void *jac(), int *mf);
#else
lsode(void (*)(const int*, double *, double*, double*), int *, double *, double *,
      double *, int *, double *,
      double *, int *, int *,
      int *, double *, int *,
      int *, int *, void (*)(), int *);
#endif

const char *
LsodeIntegrator::documentation() const
{
  return
    "This is the March 30, 1987 version of\n"
    "lsode.. Livermore Solver for Ordinary Differential Equations.\n"
    "\n"
    "Lsode is a package based on the Gear and Gearb packages, and on the\n"
    "October 23, 1978 version of the tentative Odepack user interface\n"
    "standard, with minor modifications.\n"
    "-------------------------------------------------\n"
    "Reference:\n"
    "  Alan C. Hindmarsh,  Odepack, A Systematized Collection Of Ode\n"
    "  Solvers, in Scientific Computing, R. S. Stepleman et al. (eds.),\n"
    "  North-Holland, Amsterdam, 1983, pp. 55-64.\n"
    "-------------------------------------------------\n"
    "Author and contact:  Alan C. Hindmarsh,\n"
    "                     Computing And Mathematics Research Div., L-316\n"
    "                     Lawrence Livermore National Laboratory\n"
    "                     Livermore, CA 94550.\n"
    "-------------------------------------------------";
}

bool
LsodeIntegrator::call_integrator()
{
  lsode(&hairy_f, &neqn, current_system.state_vector,
	systemtime, &tout, &itol, relerror, abserror,
	&itask, &istate, &iopt, rwork, &lrw, iwork, &liw,
	NULL, &mf);
  return true;
}

bool
LsodeIntegrator::make_setup(int oldmf)
{
  if (unsigned(neqn) != current_system.number_of_states || oldmf != mf) {
    neqn = current_system.number_of_states;
    lrw = mf == 10 ? 20 + 16*neqn :
          mf == 22 ? 22 +  9*neqn + neqn*neqn : 0;
    delete [] rwork;
    rwork = new double[(unsigned) lrw];
    delete [] iwork;
    liw = 20 + (mf != 10)*neqn;
    iwork = new int[liw];
  }
  return rwork && iwork;
}
