/* -*- c++ -*-
Copyright  1993 Bengt Mrtensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: GnansSystem.h,v 1.2 1994/01/30 20:50:11 bengt Exp $

#ifndef _GnansSystem_h
#define _GnansSystem_h

#include "Vect.h"

class GnansSystem
{
public:
  // The following guys contain the number of states (etc) in the currently
  // active system
  unsigned int number_of_states;
  unsigned int number_of_state_vectors;
  unsigned int number_of_state_matrices;
  unsigned int number_of_parameters;
  unsigned int number_of_parameter_vectors;
  unsigned int number_of_parameter_matrices;
  unsigned int number_of_functions;
  unsigned int number_of_function_vectors;
  unsigned int number_of_function_matrices;
  unsigned int number_of_randoms;
  unsigned int number_of_uniforms;
  unsigned int number_of_normals;
  unsigned int number_of_wieners;

  unsigned int state_rows;	// # rows if state a matrix 
  unsigned int state_cols;	// # cols if state a matrix 

  // The type of the currently active system 
  SystemType system_type;

  // The name of the currently active system 
  std::string system_name;

  // State vector for the integration 
  Vect state_vector;

  Vect *wiener_coeffs;

  // Pointers to functions declared in the system 
  // gnans_init_ptr is a pointer to gnans_init etc.
  int (*init_ptr)(double y[], Continuation continuation, bool stochastic);
  int (*setup_ptr)(double **x, double *t);
  void (*finish_ptr)();
  void (*f_ptr)(double *t, double *x, double *dx);
  void (*discrete_f_ptr)();
  void (*f_accepted_ptr)();
  void (*f_store_ptr)();
  void (*destroy_ptr)();

  // Arrays of states etc.
  UserParameterPointer *states;
  UserParameterPointer *parameters;
  UserParameterPointer *functions;
  UserParameterPointer *randoms;
  UserParameterPointer *wieners;

  GnansSystem() { states = parameters = functions = randoms = wieners = NULL; }
};

#endif
