// $Id: Resource.cc,v 1.4 1994/11/01 12:52:38 bengt Exp bengt $

#ifdef __GNUG__
#pragma implementation
#endif

#include <config.h>
#include "externals.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#ifdef USE_PRUSAGE_IRIX
#define USE_PRUSAGE
#define pr_utime pu_utime
#define pr_stime pu_stime
#endif
#include "Resource.h"
#ifdef USE_GETRUSAGE
#include <sys/resource.h>
#endif

unsigned int 
Resource::kilobytes() const
{
  return int(8*trajectories_current_size * number_of_stores/1000);
}

#if defined(USE_GETRUSAGE) && defined(USE_PRUSAGE)
#error Must not define both USE_GETRUSAGE and USE_PRUSAGE
#endif

#ifdef USE_PRUSAGE

// Inefficient to open anew?
Resource::Resource() {
  sprintf(proc,"/proc/%d", int(getpid()));
  if ((fd = open(proc, O_RDONLY)) == -1)
    gnans_error(impossible, "Error by opening proc");
  if (ioctl(fd, PIOCUSAGE, &prusage) == -1)
    gnans_error(impossible, "Error in ioctl in proc");
}

double
Resource::user_seconds() const
{
  return prusage.pr_utime.tv_sec + 1e-9*prusage.pr_utime.tv_nsec;
}

double
Resource::system_seconds() const
{
  return prusage.pr_stime.tv_sec + 1e-9*prusage.pr_stime.tv_nsec;
}

Resource::~Resource()
{
  close(fd);
}

#endif // USE_PRUSAGE

#ifdef USE_GETRUSAGE

unsigned int Resource::pages2kBytes = 0;

Resource::Resource() {
  if (pages2kBytes == 0)
    pages2kBytes = (int) (getpagesize()/1000);
  if (getrusage(RUSAGE_SELF, &Rusage))
    gnans_error(impossible, "Error by getrusage");
}

double
Resource::user_seconds() const
{
  return Rusage.ru_utime.tv_sec + 1e-6*Rusage.ru_utime.tv_usec;
}

double
Resource::system_seconds() const
{
  return Rusage.ru_stime.tv_sec + 1e-6*Rusage.ru_stime.tv_usec;
}

Resource::~Resource()
{
}

#endif // USEGETRUSAGE

#if !defined(USE_GETRUSAGE) && !defined(USE_PRUSAGE)
// Last resort: nothing works on this system. 
// But since this functionallity is not too important, we don't let 
// this stop us from running Gnans.

Resource::Resource()
{
}

double
Resource::user_seconds() const
{
  return -1;
}

double
Resource::system_seconds() const
{
  return -1;
}

Resource::~Resource()
{
}
#endif
