/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: Vect.cc,v 1.2 1994/08/19 18:38:31 bengt Exp $ */

#ifdef __GNUG__
#pragma implementation
#endif

#if !defined(STAND_ALONE) && !defined(MAIN)
#define GNANS
#include "externals.h"
#include "gnanslib.h"

#else				// ! GNANS
#include <builtin.h>
#include <stream.h>
#endif				// ! GNANS

#include "Vect.h"

void
Vect::resize(unsigned int n)
{
  if (n != size) {
    delete [] base;
    base = new double[n];
  }
  size = n;
}

void
Vect::setsize(unsigned int n)
{
  size = n;
}
#if 0
double&
Vect::operator[](unsigned int i)
{ 
  if (i>=size)
    gnans_error(impossible, "Wrong index, = %d", i);
  return base[i];
}
#endif

void
Vect::print(const char *mess) const
{
  switch (size) {
#ifdef GNANS
  case 0:
    gnans_error(debug, "%s n = 0!!", mess);
    break;
  case 1:
    gnans_error(debug, "%s n = 1; %g", mess, base[0]);
    break;
  case 2:
    gnans_error(debug, "%s n = 2; %g\t%g", mess, base[0], base[1]);
    break;
  case 3:
    gnans_error(debug, "%s n = 3; %g\t%g\t%g", mess, base[0], base[1], base[2]);
    break;
#endif
  default:
    for (unsigned i = 0; i < size; i++)
#ifdef GNANS
      gnans_error(debug, "%s n = %d; \t%g", mess, size, base[i]);
#else
    cout << base[i] << "\n";
#endif
  }
}

void
Vect::print() const
{
  print("");
}

GNU_CONST double
norm_sqr(const Vect &v)
{
  double sum = 0;
  for (unsigned int i = 0; i < v.size; i++)
    sum += v.base[i]*v.base[i];
  return sum;
}

GNU_CONST double
norm1(const Vect &v)
{
  double sum = 0;
  for (unsigned int i = 0; i < v.size; i++)
    sum += abs(v.base[i]);
  return sum;
}

void
Vect::barf(char *message)
{
#ifdef GNANS
  gnans_error(impossible, "%s", message);
#else
  cout << message << "\n";
#endif
}

#ifdef MAIN
main()
{
  double *x = new double[2];
  double *y = new double[3];
  double a = 3.14;

  x[0] = 1;
  x[1] = 2;
  y[0] = 3.14;
  y[1] = 2.71;

  cout << "Hellow world\n";
  Vect X(x, 2, 1);
  Vect Y(y, 2, 1);
  Vect Z(2);
  X.print();
  Y.print();
  Z = X+Y;
  Z.print();
  Z = X * a;
  Z.print();
  Z = a * X;
  Z.print();
  Z += X;
  Z.print();
  cout << ((double *) Z)[1] << "\n";
}
#endif
