/*
 Copyright (C) 1992-1994 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: callbacks.cc,v 1.22 1994/08/24 09:58:29 bengt Exp $ */

/*
 * Callback functions, called from the main body are declared here, together
 * with a large number of support routines
 */

#include "externals.h"
#include "CIntegrator.h"
#include "utillib.h"
#include "PlotterControl.h"

#include <assert.h>

#define String XtString
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#ifdef USEXAW3DINCLUDES

#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Form.h>
#include <X11/Xaw3d/Cardinals.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Dialog.h>
#include <X11/Xaw3d/Toggle.h>
#include <X11/Xaw3d/AsciiText.h>

#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/Sme.h>
#include <X11/Xaw3d/SmeBSB.h>
#include <X11/Xaw3d/SmeLine.h>

#else

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>

#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#endif

#include <X11/bitmaps/xlogo16>
#undef String

#ifdef MAX
#undef MAX
#endif
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#ifdef MIN
#undef MIN
#endif
#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define STR_EQUAL(s1, s2) strcmp((s1), (s2)) == 0

#define log if (do_logging) va_log_event
#define widgetlog if (do_logging && trace_widgets) va_log_event
#define MAYBE_ZAP(w) if (w) maybe_zap(&w)

// const unsigned int MAXSTRINGSIZE = 255;

/**************** Forward declarations ****************/
static Dimension width_of_display_and_change;
static void change_value(Widget mother_widget, XtPointer uservar,
			 XtPointer call_data);
static void display_and_change_var(UserParameter *uservar, bool writeable,
				   Widget parent_widget);
static void Variable_Group(UserParameter *vector[], int size,
			   bool writeable, Widget mother_widget,
			   Widget *popup_ptr, char *name);
static void update_labels(UserParameter *uservar);
static void popup_describe(Widget mother_widget, XtPointer string,
			   XtPointer call_data);
static void popup_describe_algorithm(Widget mother_widget, XtPointer junk,
				     XtPointer call_data);
static void popup_describe_format(Widget mother_widget, XtPointer junk,
				  XtPointer call_data);
static void popup_toggle(Widget parent_widget, XtPointer uservar_ptr,
			 XtPointer call_data);
static void popup_inquiry(Widget mother_widget, XtPointer uservar_ptr,
			  XtPointer call_data);
static void switch_algorithm(Widget mother_widget, XtPointer algorithm,
			     XtPointer call_data);
static void toggle_global_switch(Widget mother_widget, XtPointer client_data,
				 XtPointer call_data);
static void format_menu_select(Widget w, XtPointer number, XtPointer garbage);
static UserParameter *get_userparameter(Widget w);
static UserParameter *search_4_userparameter(Widget w,
					      UserParameter *table[],
					      unsigned int size);
static void Internal_parameters(Widget mother_widget, XtPointer client_data,
				XtPointer call_data);
static void Debug(Widget mother_widget, XtPointer client_data,
		  XtPointer call_data);
static void Breakpoint(Widget mother_widget, XtPointer client_data,
		       XtPointer call_data);
#ifdef HELP
static void Help(Widget mother_widget, XtPointer client_data,
		 XtPointer call_data);
#endif
static bool confirmation(Widget where_from, char *message);
static void maybe_zap(Widget *wp);
static void start_decode_commands(char *filename);
static void make_stdin_message(char *filename);
static char *make_popup_title(char *, const UserParameter *);
static void yes_response(Widget w, XtPointer client, XtPointer call);
static void  no_response(Widget w, XtPointer client, XtPointer call);
#ifdef DYNAMIC_LOADING
static void make_compilation_message(char *filename);
static void start_setup_system_dynamic(char *filename);
#endif

const int MAX_NO_OF_ALGOS = 20;
static Widget algo_widgets[MAX_NO_OF_ALGOS]; // Gross...

/****************************************************************/
/*
 * Callbacks for all the main buttons follow
 */
/****************************************************************/
extern void 
Start(Widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  widgetlog("Start");
  reset_cpu();
  integrate(init_all);
}

/****************************************************************/
extern void 
Stop(Widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  if (running_script) {
    gnans_error(message, "Stop button pressed, stopping script execution");
    log("Stop button pressed, stopping script execution");
  } else {
    if (running_integration) {
      gnans_error(message, "Stop button pressed at time %g",
		  double(systemtime));
      log("Stop button pressed at time %g", double(systemtime));
    }
  }
  running_integration = false;
  abort_script = running_script;
}

/****************************************************************/
#ifdef DYNAMIC_LOADING
// The kludges in this and the following function is due to the fact
// that the braindamaged fileselector does not understand String's, and
// it writes to them!!
// ** Remove this uglyness as soon as possible!!!!!!!!!!!!

static char system_select_kludge[MAXSTRINGSIZE];
extern void 
System(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  widgetlog("System");
  strcpy(system_select_kludge, system_filename.c_str());
  fileselector("System Select", "system_select", ".gnans",
	       system_select_kludge, &start_setup_system_dynamic,
	       &make_compilation_message, true, false);
}
#endif /* DYNAMIC_LOADING */

/****************************************************************/
static char script_select_kludge [MAXSTRINGSIZE];
extern void
Script(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  widgetlog("Script");
  strcpy(script_select_kludge, scriptfilename.c_str());
  fileselector("Script Select", "script_select", "",
	       script_select_kludge, &start_decode_commands,
	       &make_stdin_message,
	       true, true);
}

/****************************************************************/
extern void 
Display_vars_callback(Widget, XtPointer /*client_data*/,
		      XtPointer /*call_data*/)
{
  widgetlog("Display_vars");
  if (do_logging)
    display_in_logfile();
  Display_vars(0);
}

/****************************************************************/
extern void
Continue(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  widgetlog("Continue");
  reset_cpu();
  integrate(init_none);
}

/****************************************************************/
/* Should be fancier */

/*
 * If called with first argument NULL (i.e. not as a callback but directly),
 * just write on stdout.
 */
extern void
About(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  bool Xsaved = X;
  if (mother_widget) {
    widgetlog("About");
    gnans_error(nonerror_message,
		"Gnans Version %s.\n"
		"Compiled on %s.\nConfigured with command:\n%s\n\n"
		"Copyright \251 1992, 1993, 1994, 1995, 1996 Bengt M\345rtensson and "
		"Institute for Dynamical Systems\n\n"

#define RELEASED_VERSION
#ifdef RELEASED_VERSION
"Gnans is free software; you can redistribute it and/or modify it under\n"
"the terms of the GNU General Public License as published by the Free\n"
"Software Foundation; either version 2, or (at your option) any later\n"
"version.\n\n"

"Gnans is distributed in the hope that it will be useful, but WITHOUT\n"
"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or\n"
"FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License\n"
"for more details.\n\n"

"You should have received a copy of the GNU General Public License\n"
"along with Gnans; see the file COPYING.  If not, write to the \n"
"Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.\n\n"
#else

"This version is an experimental development version. Released\n"
"versions are released under the GNU General Public License,\n"
"however, this version may not be redistributed without the consent\n"
"of the author.\n\n"
#endif

"Gnans is documented in the Texinfo format, which allows the same file\n"
"to generate both an online, hypertext help system, and a version which\n"
"can be printed with TeX.  Please consult this file, either reading it\n"
"online with, e.g. Gnu Emacs or another info reader, or read the\n"
"printed documentation.\n\n"


"       Comments, bugreports, bugfixes, gripes, enhancements, etc to\n"
"       bengt@math.uni-Bremen.de.",
              version_string, compiled_string, compilation_characteristics);
  }
  else {
    X = false;      
    gnans_error(nonerror_message,
		"Gnans Version %s.\n"
		"Compiled on %s.\nConfigured with command:\n%s\n\n"
		"Copyright \251 1992, 1993, 1994, 1995, 1996 Bengt M\345rtensson and\n"
		"Institute for Dynamical Systems\n\n"

"Gnans is documented in the Texinfo format, which allows the same file\n"
"to generate both an online, hypertext help system, and a version which\n"
"can be printed with TeX.  Please consult this file, either reading it\n"
"online with, e.g. Gnu Emacs or another info reader, or read the\n"
"printed documentation.\n\n"


"       Comments, bugreports, bugfixes, gripes, enhancements, etc to\n"
"       bengt@math.uni-Bremen.de.",
		version_string, compiled_string, compilation_characteristics);
  }
  X = Xsaved;
}

#ifdef HELP
/****************************************************************/
extern void
Help(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  widgetlog("Help");
  help();
}
#endif

/****************************************************************/
extern void
Debug(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  log("Coredump button pressed");
  gnans_error(coredump, "Coredump button pressed");
}

/****************************************************************/
extern void
Breakpoint(Widget /*mother_widget*/, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  log("Breakpoint");
}

/****************************************************************/
extern void 
Quit(Widget widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  log("Quit");
  if (confirmation(widget, "Please confirm quit"))
    die("", 0);
}

/****************************************************************/
/*
 * These widgets are global (for the file), in order to pop down
 * from other functions. A functions destroying such a widget should
 * NULL the corresponding Widget.
 */

static Widget parameter_popup = NULL;
static Widget states_popup = NULL;
static Widget functions_popup = NULL;
static Widget seeds_popup = NULL;
static Widget times_popup = NULL;
static Widget settings_popup = NULL;
static Widget internal_parameters_popup = NULL;

/****************************************************************/
extern void 
Parameters(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Variable_Group(current_system.parameters,
		 current_system.number_of_parameters, true,
		 mother_widget, &parameter_popup, "Parameters");
}

/****************************************************************/
extern void 
Functions(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Variable_Group(current_system.functions,
		 current_system.number_of_functions, false,
		 mother_widget, &functions_popup, "Functions");
}

/****************************************************************/
extern void 
Seeds(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Variable_Group(current_system.randoms,
		 current_system.number_of_randoms, true,
		 mother_widget, &seeds_popup, "Seeds");
}

/****************************************************************/
extern void 
States(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Variable_Group(current_system.states, current_system.number_of_states,
		 true, mother_widget, &states_popup, "States");
}

/****************************************************************/
extern void 
Times(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Widget box, button;
  char title[MAXSTRINGSIZE];

  widgetlog("Times");
  MAYBE_ZAP(times_popup);
  make_popup_title(title, "Times");
  times_popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
				     mother_widget, XtNtitle, title, NULL);
  box = XtVaCreateManagedWidget("box", boxWidgetClass, times_popup, NULL);
  display_and_change_var(&systemtime, false, box);
  display_and_change_var(&starttime, true, box);
  display_and_change_var(&stoptime, true, box);
  if (current_system.system_type != discrete)
    display_and_change_var(&timestep, true, box);
  display_and_change_var(&storeincrement, true, box);
  if (current_system.number_of_wieners)
    display_and_change_var(&wienertick, true, box);
  button =  XtVaCreateManagedWidget("ok", commandWidgetClass, box, NULL);
  XtAddCallback(button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &times_popup);
  popup_position(times_popup, mother_widget, XtGrabNone, false);
}

////////////////////////////////////////////////////////////////
// This function activates the plotter and presents a control panel.
extern void 
Plot(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  if (number_of_stores == 0) {
    gnans_error(non_fatal, "Nothing to plot");
    return;
  };

  widgetlog("Plot");
  (void) new PlotterControl(mother_widget);
}

/****************************************************************/
static Pixmap mark;

extern void
Settings(Widget mother_widget, XtPointer /*client_data*/, XtPointer /*call_data*/)
{
  Widget box, button, menu, entry;
  Widget button_debug;
  unsigned int i;
  char title[MAXSTRINGSIZE];

  widgetlog("Settings");
  MAYBE_ZAP(settings_popup);
  make_popup_title(title, "Gnans/Settings");
  settings_popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
					mother_widget, XtNtitle, title,
					NULL);
  box = XtVaCreateManagedWidget("box", boxWidgetClass, settings_popup, NULL);

  /* Create the bitmap for marking selected items. */
  mark = XCreateBitmapFromData(XtDisplay(toplevel),
			       RootWindowOfScreen(XtScreen(toplevel)),
			       (const char *) xlogo16_bits, xlogo16_width,
			       xlogo16_height);
  /*algorithms */
  button = XtVaCreateManagedWidget("algorithm_menu_button",
				   menuButtonWidgetClass, box,
				   XtNsensitive,
				   current_system.system_type == continuous,
				   NULL);
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass,
			      button, NULL);
  entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu, NULL);
  for (i = 0; algorithms[i]; i++) {
    entry = XtVaCreateManagedWidget(algorithms[i]->name(),
				    smeBSBObjectClass, menu, 
				    XtNleftBitmap,
				    current_algorithm == algorithms[i]
				    ? mark : None,
				    XtNsensitive,
				    (! current_system.number_of_wieners
				     || algorithms[i]->stochastic())
				    || expert_mode,
				    NULL);
    algo_widgets[i] = entry;
    XtAddCallback(entry, XtNcallback, switch_algorithm, (XtPointer) i);
  }

  /* formats */
  button = XtVaCreateManagedWidget("format_menu_button", menuButtonWidgetClass,
				   box, NULL);
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass,
			      button, NULL);
  entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu, NULL);
  for (i = 0; store_format_table[i].name; i++) {
    entry = XtVaCreateManagedWidget(store_format_table[i].name,
				    smeBSBObjectClass, menu, 
				    XtNleftBitmap,
				    i == store_format ? mark : None,
				    NULL);
    store_format_table[i].widget = entry;
    XtAddCallback(entry, XtNcallback, format_menu_select, (XtPointer) i);
  }

  /* switches */
  button = XtVaCreateManagedWidget("switches_menu_button",
				   menuButtonWidgetClass, box,
				   NULL);
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass,
			      button, NULL);
  entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu, NULL);
  for (i = 0; switches_table[i].ptr; i++) {
    entry = XtVaCreateManagedWidget(switches_table[i].name,
				    smeBSBObjectClass, menu, 
				    XtNleftBitmap,
				    *switches_table[i].ptr ? mark : None,
				    NULL);
    switches_table[i].widget = entry;
    XtAddCallback(entry, XtNcallback, toggle_global_switch,
		  (XtPointer) &(switches_table[i]));
  }

  /* parameters */
  button = XtVaCreateManagedWidget("internal_parameters",
				   commandWidgetClass, box, NULL);
  XtAddCallback(button, XtNcallback, Internal_parameters, NULL);


  button =  XtVaCreateManagedWidget("describe_algorithm", commandWidgetClass,
				    box,
				    XtNsensitive,
				    current_system.system_type != discrete,
				    NULL);
  XtAddCallback(button, XtNcallback, popup_describe_algorithm, NULL);
  button =  XtVaCreateManagedWidget("describe_format", commandWidgetClass,
				    box, NULL);
  XtAddCallback(button, XtNcallback, popup_describe_format, NULL);

  if (development) {
    button_debug = XtVaCreateManagedWidget("debug", commandWidgetClass,
					   box, NULL);
    XtAddCallback(button_debug, XtNcallback, Debug, NULL);

    Widget btn = XtVaCreateManagedWidget("breakpoint", commandWidgetClass,
					 box, NULL);
    XtAddCallback(btn, XtNcallback, Breakpoint, NULL);
  }

  button =  XtVaCreateManagedWidget("ok", commandWidgetClass, box, NULL);
  XtAddCallback(button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &settings_popup);
  popup_position(settings_popup, mother_widget, XtGrabNone, false);
}

/****************************************************************/
/*
 * Callback invoked when the user selects an algorithm from the menu
 */
static void 
switch_algorithm(Widget /*mother_widget*/, XtPointer algorithm_no,
		 XtPointer /*call_data*/)
{
  int i;
  select_algorithm((unsigned int) algorithm_no);
  for (i = 0; algorithms[i]; i++)
    XtVaSetValues(algo_widgets[i],
		  XtNleftBitmap, algorithms[i] == current_algorithm ?
		  mark : None,
		  NULL);
}

/****************************************************************/
/*
 * Callback invoked when the user selects a store format
 */
static void
format_menu_select(Widget /*w*/, XtPointer number, XtPointer /*garbage*/)
{
  unsigned int i;
  store_format = (int) number;
  write_trajectories();
  widgetlog("format_menu_select %s", store_format_table[store_format].name);
  for (i = 0; store_format_table[i].name; i++) {
    XtVaSetValues((Widget) store_format_table[i].widget,
		  XtNleftBitmap, i == store_format ? mark : None,
		  NULL);
  }
}

/****************************************************************/
/*
 * Callback invoked when the user selects a switch to be toggled
 */
static void
toggle_global_switch(Widget /*mother_widget*/, XtPointer s, XtPointer /*call_data*/)
{
  Switch *sw = (Switch *) s;
  log("toggle_global_switch (%s); %s -> %s",
      (char *) sw->name,
      *sw->ptr ? "true"  : "false",
      *sw->ptr ? "false" : "true");
  *sw->ptr = ! *sw->ptr;
  XtVaSetValues((Widget) sw->widget, XtNleftBitmap, *sw->ptr ? mark : None, NULL);
  if (sw->fcn)
    (*sw->fcn)();
}

/****************************************************************/
static void
Internal_parameters(Widget mother_widget, XtPointer /*client_data*/,
		    XtPointer /*call_data*/)
{
  Widget box, button;
  char str[MAXSTRINGSIZE];

  widgetlog("Internal_parameters");
  MAYBE_ZAP(internal_parameters_popup);
  make_popup_title(str, "Internal Parameters");
  internal_parameters_popup =
    XtVaCreatePopupShell("popup",
			 transientShellWidgetClass,
			 mother_widget,
			 XtNtitle, str,
			 NULL);
  box = XtVaCreateManagedWidget("box", boxWidgetClass,
				internal_parameters_popup, NULL);
  if (current_system.system_type != discrete) {
    display_and_change_var(&abserror, true, box);
    display_and_change_var(&relerror, true, box);
    if (expert_mode || development)
      display_and_change_var(&method_flag, true, box);
    if (development) {
      display_and_change_var(&algorithm_parameter_1, true, box);
      display_and_change_var(&algorithm_parameter_2, true, box);
      display_and_change_var(&algorithm_parameter_3, true, box);
    }
  }
  button =  XtVaCreateManagedWidget("ok", commandWidgetClass, box, NULL);
  XtAddCallback(button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &internal_parameters_popup);
  popup_position(internal_parameters_popup, mother_widget, XtGrabNone, false);
}

/****************************************************************/
static void 
popup_describe_algorithm(Widget mother_widget, XtPointer /*junk*/,
			 XtPointer call_data)
{
  widgetlog("popup_describe_algorithm");
  popup_describe(mother_widget,
		 (XtPointer) current_algorithm->documentation(), call_data);
}

/****************************************************************/
static void 
popup_describe_format(Widget mother_widget, XtPointer /*junk*/,
		      XtPointer call_data)
{
  widgetlog("popup_describe_format");
  popup_describe(mother_widget,
		 (char *) store_format_table[store_format].documentation,
		 call_data);
}

/****************************************************************/
/****************************************************************/
/****************************************************************/
/*
 * Support routine for States() etc
 * Allows the user to inspect and possibly change an array of UserParameter
 */
static void 
Variable_Group(UserParameter *vector[], int size,
	       bool writeable, Widget mother_widget,
	       Widget *popup_ptr, char *name)
{
  Widget popup, box, button;
  int i;
  char title[MAXSTRINGSIZE];

  widgetlog(name);
  maybe_zap(popup_ptr);
  make_popup_title(title, name);
  popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
			       mother_widget, XtNtitle, title, NULL);
  *popup_ptr = popup;
  box = XtVaCreateManagedWidget("box", boxWidgetClass, popup, NULL);
  for (i = 0; i < size; i++)
    display_and_change_var(vector[i], writeable, box);

  button =  XtVaCreateManagedWidget("ok", commandWidgetClass,
				    box, NULL);
  XtAddCallback(button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) popup_ptr);
  popup_position(popup, mother_widget, XtGrabNone, false);
}

/*
 * Support routine for Variable_Group()
 * Allows for the inspection and possible changing of a UserParameter
 */
static void
display_and_change_var(UserParameter *uservar, bool writeable,
		       Widget parent_widget)
{
  Widget box, button, label, innerbox;
  Dimension w;
  
  widgetlog("display_and_change_var(``%s'')", uservar->name());
  box = XtVaCreateManagedWidget(uservar->name(), boxWidgetClass, parent_widget,
				NULL);
  label = XtVaCreateManagedWidget("label", labelWidgetClass, box, NULL);
  uservar->widget = label;
  
  innerbox = XtVaCreateManagedWidget("innerbox", boxWidgetClass, box,
				     XtNorientation, XtorientHorizontal,
				     NULL);

  XtVaGetValues(innerbox, XtNhSpace, &w, NULL);
  width_of_display_and_change = 4*w;
  XtVaGetValues(innerbox, XtNborderWidth, &w, NULL);
  width_of_display_and_change += 2*w;
  button = XtVaCreateManagedWidget("change", commandWidgetClass, innerbox,
				   XtNsensitive, int(writeable), NULL);
  XtVaGetValues(button, XtNwidth, &w, NULL);
  width_of_display_and_change += w;
  XtAddCallback(button, XtNcallback, popup_inquiry, (XtPointer) uservar);

  button = XtCreateManagedWidget("describe", commandWidgetClass, innerbox,
				 (ArgList) NULL, ZERO);
  XtVaGetValues(button, XtNwidth, &w, NULL);
  width_of_display_and_change += w;
  XtAddCallback(button, XtNcallback, popup_describe,
		(XtPointer) (uservar->documentation()));
  button = XtCreateManagedWidget("toggle", commandWidgetClass, innerbox,
				 (ArgList) NULL, ZERO);
  XtVaGetValues(button, XtNwidth, &w, NULL);
  width_of_display_and_change += w;
  XtVaGetValues(button, XtNborderWidth, &w, NULL);
  width_of_display_and_change += 6*w;
  XtAddCallback(button, XtNcallback, popup_toggle, (XtPointer) uservar);
  update_labels(uservar);
}

// Callback for display_and_change_var()
// Allows the change of a user_variable
static void 
popup_inquiry(Widget mother_widget, XtPointer uservar_ptr, XtPointer /*call_data*/)
{
  Widget inquiry_popup, dialog, ok_button, cancel_button;
  char title[MAXSTRINGSIZE];
  UserParameter *user_par = (UserParameter *) uservar_ptr;
  
  make_popup_title(title, user_par);
  widgetlog("popup_inquiry %s", user_par->name());
  inquiry_popup = XtVaCreatePopupShell("inquirypopup",
				       transientShellWidgetClass,
				       mother_widget,
				       XtNtitle, title,
				       NULL);
  char s[MAXSTRINGSIZE];
  sprintf(s, "%g", double(*user_par));
  dialog = XtVaCreateManagedWidget("dialog", dialogWidgetClass, inquiry_popup,
				   XtNvalue, (char *) s, NULL);
  
  ok_button =  XtCreateManagedWidget("ok", commandWidgetClass, dialog,
				     (ArgList) NULL, 0);
  XtAddCallback(ok_button, XtNcallback, change_value, uservar_ptr);
  cancel_button =  XtCreateManagedWidget("cancel", commandWidgetClass,
					 dialog, (ArgList) NULL, 0);
  XtAddCallback(cancel_button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &inquiry_popup);
  XtInstallAccelerators(dialog, ok_button);
  popup_position(inquiry_popup, mother_widget, XtGrabNone, true);
}

/*
 * Callback for display_and_change_var()
 * Pops up the description for a user_variable
 */
static void
popup_describe(Widget mother_widget, XtPointer string, XtPointer /*call_data*/)
{
  Widget popup, dismiss, box;
  char str[MAXSTRINGSIZE], title[MAXSTRINGSIZE];

  strncpy(str, (char *) string, 50);
  if (strlen((char *) str) >= 50) {
    str[50] = '\0';
    strcat(str, "...");
  }
  widgetlog("popup_describe(%s)", str);

  make_popup_title(title, "Description");
  popup = XtVaCreatePopupShell("popup", transientShellWidgetClass,
			       mother_widget, XtNtitle, title, NULL);
  box = XtVaCreateManagedWidget("box", boxWidgetClass, popup, NULL);
#ifdef USETEXTWIDGET
  XtVaCreateManagedWidget("description", asciiTextWidgetClass, box,
				 XtNstring, (char *) string,
				 NULL);
#else
  XtVaCreateManagedWidget("label", labelWidgetClass, box,
                                 XtNlabel, (char *) string, NULL);
#endif
  dismiss = XtVaCreateManagedWidget("dismiss", commandWidgetClass,
				    box, NULL);
  XtAddCallback(dismiss, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &popup);
  popup_position(popup, mother_widget, XtGrabNone, false /* true */);
}

/*
 * Callback for display_and_change_var()
 * Toggles the store status of a user_variable.
 */
static void 
popup_toggle(Widget /*parent_widget*/, XtPointer uservar_ptr,
	     XtPointer /*call_data*/)
{
  UserParameter *uservar = (UserParameter *) uservar_ptr;
  widgetlog("popup_toggle", uservar->name());
  uservar->toggle_store();
  update_labels(uservar);
  if (append_trajectories)
    write_trajectories();
  storeset_dirty = true;
}

/*
 * Callback invoked from popup_inquiry, when the user presses Ok
 */
static void 
change_value(Widget mother_widget, XtPointer uservar, XtPointer /*call_data*/)
{
  Widget w;
  Widget thedialogwidget = XtParent(mother_widget);
  double value = 0;
  UserParameter *the_param = (UserParameter *) uservar;
  if (sscanf((char *) XawDialogGetValueString(thedialogwidget), "%lg", &value)
      > 0) {
    log("change_value %s: %g -> %g",
	the_param->name(), double(*the_param), value);
    *the_param = value;
  }
  update_labels(the_param);
  XtVaSetValues(XtParent(XtParent(XtParent(mother_widget))),
		XtNsensitive, True, NULL);
  w = XtParent(thedialogwidget);
  XtDestroyWidget(w);
  w = NULL;
}

/*
 * Update the labels for a UserParameter
 */
static void 
update_labels(UserParameter *uservar)
{
  char widgetname_and_value[MAXSTRINGSIZE];
  sprintf(widgetname_and_value,"%s:  %g (%sstored)", uservar->name(),
		 double(*uservar), uservar->store_me() ? "" : "Not ");
  XtVaSetValues((Widget) uservar->widget,
		XtNlabel, widgetname_and_value,
		XtNwidth, width_of_display_and_change, NULL);
}

/****************************************************************/
/*
 * Support routine to Script()
 * If the argument is "", informs the user that commands are now taken
 * from stdin.
 */

/* The usual static widget */
static Widget stdin_message = NULL;

static void
make_stdin_message(char *filename)
{
  Widget label;
  char title[MAXSTRINGSIZE];

  if (STR_EQUAL(filename, "")) {
    make_popup_title(title, "Stdin Message");
    stdin_message = XtVaCreatePopupShell("stdin_message",
					 transientShellWidgetClass,
					 toplevel, XtNtitle, title, NULL);
    label = XtVaCreateManagedWidget("label", labelWidgetClass, stdin_message,
				    NULL);
    popup_position(stdin_message, toplevel, XtGrabExclusive, true);
  }
}

static void
destroy_stdin_message()
{
  if (stdin_message)
    XtDestroyWidget(stdin_message);
  
  /*
   * Throw away all events in the queue
   * Corresponds to disabling "mouse-ahead" while we are reading stdio
   */
  XFlush(XtDisplay(toplevel));
  XSync(XtDisplay(toplevel), True);
  stdin_message = NULL;
}

/* Support routine to Script(). Starts the work. */
static void
start_decode_commands(char *filename)
{
  assert(filename);
  widgetlog("start_decode_commands(%s)", filename);
  scriptfilename = filename;
  decode_commands(filename);
  destroy_stdin_message();
}

/****************************************************************/
#ifdef DYNAMIC_LOADING
/* Support stuff for System() */

static Widget compilation_message = NULL;

static void
make_compilation_message(char */*filename*/)
{
  Widget label;
  char title[MAXSTRINGSIZE];

  make_popup_title(title, "Compilation Message");
  compilation_message = XtVaCreatePopupShell("compilation_message",
					     transientShellWidgetClass,
					     toplevel, XtNtitle, title,
					     NULL);
  label = XtVaCreateManagedWidget("label", labelWidgetClass,
				  compilation_message, NULL);
  popup_position(compilation_message, toplevel, XtGrabExclusive, true);
}

static void
destroy_compilation_message()
{
  XtDestroyWidget(compilation_message);
  compilation_message = NULL;
}

static void
start_setup_system_dynamic(char *filename)
{
  assert(filename);
  widgetlog("start_setup_system_dynamic(%s)", filename);
  system_filename = filename;
  enable_busy_window();
  setup_system_dynamic(std::string(filename));
  disable_busy_window();
  destroy_compilation_message();
}
#endif /* DYNAMIC_LOADING */

/****************************************************************/
/*
 * General purpose callback for zapping global popups widgets
 */
extern void 
destroy_inquiry_popup(Widget mother_widget, XtPointer null_me,
		      XtPointer /*call_data*/)
{
  Widget w = XtParent(XtParent(mother_widget));
  XtVaSetValues(XtParent(w), XtNsensitive, True, NULL);
  XtDestroyWidget(w);
  if (null_me)
    *((Widget *) null_me) = NULL;  
}

/*
 * Simple callback corresponding to Cancel
 */
static void 
destroy_popup(Widget, XtPointer zap_me, XtPointer /*call_data*/)
{
  XtDestroyWidget((Widget) zap_me);
}

/****************************************************************/
/*
 * Positions and pops up a popup.
 */

extern void 
popup_position(Widget popup, Widget parent, XtGrabKind grab_type,
	       bool disable_parent)
{
#define WM_HEAD_HEIGHT 22
  Position x, y;
  Dimension width, height, border_width;

  Window root, child;
  int dummyx, dummyy, xx, yy;
  unsigned int dummymask;

  if (disable_parent)
    XtVaSetValues(parent, XtNsensitive, False, NULL);

  XtRealizeWidget(popup);

#if POSITION_RELATIVE_TO_PARENT
  XtVaGetValues(parent,
		XtNwidth, &width,
		XtNheight, &height,
		NULL);
  XtTranslateCoords(parent,
		    (Position) width/3, (Position) height/3,
		    &x, &y);
#else
  XQueryPointer(XtDisplay(popup), XtWindow(popup), &root, &child,
		&xx, &yy,
		&dummyx, &dummyy, &dummymask);
  x = (Position) xx;
  y = (Position) yy;
#endif

  XtVaGetValues(popup,
		XtNwidth, &width,
		XtNheight, &height,
		XtNborderWidth, &border_width,
		NULL);
  width += 2*border_width;
  height += 2*border_width;
  x = (Position) MAX(x - width/2, 0);
  y = (Position) MAX(y - height/2, 0);
  x = (Position) MIN(x, XtScreen(popup)->width - width);
  y = (Position) MIN(y, XtScreen(popup)->height - WM_HEAD_HEIGHT - height);
  XtVaSetValues(popup, XtNx, x, XtNy, y, NULL);
  XtPopup(popup, grab_type);
}

/****************************************************************/
extern void 
destroy_toplevel()
{
  XtDestroyApplicationContext(XtWidgetToApplicationContext(toplevel));
}

extern void zap_popups()
{
  MAYBE_ZAP(times_popup);    
  MAYBE_ZAP(settings_popup);    
  MAYBE_ZAP(states_popup);    
  MAYBE_ZAP(functions_popup);    
  MAYBE_ZAP(seeds_popup);    
  MAYBE_ZAP(parameter_popup);
  PlotterControl::destroy_plotter_controls();
}
/****************************************************************/
/*
 * General purpose function to popup an error message
 */
extern void 
popup_error_message(Widget mother_widget, /* Position relative to */
		    XtPointer /*client_data*/, /* Not used */
		    XtPointer the_error_message, /* Error message */
		    /*XtPointer*/ const char *the_title, /* Window title */
		    XtGrabKind grab_type /* Grabbing the focus? */
		    )
{
  Widget errorpopup, dismiss, box, errormessage, errorlabel;
  char message[10*MAXSTRINGSIZE], title[MAXSTRINGSIZE];

  make_popup_title(title, (char *) the_title);
  sprintf(message, abort_script ? "%s\nScript execution aborted" : "%s",
	  (char *) the_error_message);
  errorpopup = XtVaCreatePopupShell("errorpopup", transientShellWidgetClass,
				    mother_widget, XtNtitle, title, NULL);
  box = XtVaCreateManagedWidget("box", formWidgetClass, errorpopup, NULL);
  errorlabel = XtVaCreateManagedWidget("errorlabel", labelWidgetClass, box,
				       XtNlabel, the_title, NULL);
  errormessage = XtVaCreateManagedWidget("errormessage",
					 labelWidgetClass, box,
					 XtNlabel, message, NULL);
  
  make_equal_width(errorlabel, errormessage);
  dismiss = XtCreateManagedWidget("dismiss", commandWidgetClass,
				  box, (ArgList) NULL, ZERO);
  XtAddCallback(dismiss, XtNcallback, destroy_popup, (XtPointer) errorpopup);

  popup_position(errorpopup, mother_widget, grab_type, true);
}
/****************************************************************/
/*
 * Actions, duplicating callbacks, enabling the user to press return
 * instead of clicking Ok.
 *
 * I am absolutely convinced that there is a much simpler way to do it...
 */
extern void 
change_value_action(Widget widget, XEvent */*event*/, XtString */*params*/,
		    Cardinal */*num_params*/)
{
  UserParameter *p;
  p = get_userparameter(widget);
  if (p)
    change_value(widget, (XtPointer) p, NULL);
}
  
/*
 * Support routine to change_value_action.
 * Returns the UserParameter associated with the widget w
 */
static UserParameter *
get_userparameter(Widget w)
{
  UserParameter *p;
  p = search_4_userparameter(w, current_system.states,
			     current_system.number_of_states);
  if (p)
    return p;
  p = search_4_userparameter(w, internals, number_of_internals);
  if (p)
    return p;
  p = search_4_userparameter(w, current_system.parameters,
			     current_system.number_of_parameters);
  if (p)
    return p;
  p = search_4_userparameter(w, current_system.randoms,
			     current_system.number_of_randoms);
  if (!p)
    gnans_error(non_fatal, "Internal error, did not find uservar");
  return p;
}

/*
 * Searches for a UserParameter corresponding to the widget w in
 * a UserParameter array. Returns NULL on failure.
 */
static UserParameter *
search_4_userparameter(Widget w, UserParameter *table[], unsigned int size)
{
  unsigned int i;
  for (i = 0; i < size; i++)
    if (table[i]->widget
	&& (XtParent((Widget)table[i]->widget) ==
	    XtParent(XtParent(XtParent(XtParent(XtParent(w))))))) {
      return table[i];
    }
  /* Not found in the table */
  return NULL;
}

/****************************************************************/
extern void
destroy_popup_action(Widget widget, XEvent */*event*/, XtString */*params*/,
		     Cardinal */*num_params*/)
{
  /*
   * If this widget is not called "errorpopup", call myself recursive
   * with the parent as argument
   */
  if (STR_EQUAL(XtName(widget), "errorpopup"))
    XtDestroyWidget(widget);
  else
    destroy_popup_action(XtParent(widget), NULL, NULL, NULL);
}

/****************************************************************/
/*
 * Ask the user for confirmation synchronously.
 *
 * This code is derived from X FAQ, where it is commented.
 */
static int confirm_answer;

static bool
confirmation(Widget where_from, char *message)
{
  Widget confirmpopup, box, label, yes, no;
  char title[MAXSTRINGSIZE];
  XtAppContext context;

  confirm_answer = 2;
  make_popup_title(title, "Confirm");
  confirmpopup = XtVaCreatePopupShell("confirm", transientShellWidgetClass,
				      where_from, XtNtitle, title, NULL);
  box = XtVaCreateManagedWidget("box", boxWidgetClass, confirmpopup, NULL);
  label = XtVaCreateManagedWidget("label",
				  labelWidgetClass, box,
				  XtNlabel, message, NULL);
  
  yes = XtVaCreateManagedWidget("yes", commandWidgetClass, box, NULL);
  no = XtVaCreateManagedWidget("no", commandWidgetClass, box, NULL);
  XtAddCallback(yes, XtNcallback, yes_response, NULL);
  XtAddCallback(no, XtNcallback, no_response, NULL);
  popup_position(confirmpopup, where_from, XtGrabExclusive, false);
  context = XtWidgetToApplicationContext(where_from);
  while (confirm_answer == 2 || XtAppPending(context))
    XtAppProcessEvent(context, XtIMAll);
  XtDestroyWidget(confirmpopup);
  return confirm_answer != 0;
}

/* Callback, corresponding to Yes button */
static void
yes_response(Widget /*w*/, XtPointer /*client*/, XtPointer /*call*/)
{
  confirm_answer = 1;
}
  
/* Callback, corresponing to No button */
static void
no_response(Widget /*w*/, XtPointer /*client*/, XtPointer /*call*/)
{
  confirm_answer = 0;
}
/****************************************************************/
/*
 * Misc. support functions
 */

// Format a title for a popup in a consistent way
extern char *
make_popup_title(char *title, const char *name)
{
  sprintf(title, "%s [%s]", name, current_system.system_name.c_str());
  return title;
}

static char *
make_popup_title(char *title, const UserParameter *up)
{
  return make_popup_title(title, up->name());
}

// Zap a widget, provided that it makes sense
static void 
maybe_zap(Widget *wp)
{
  /* This is probably not quite correct, sometimes segmentation faults... */
  if (*wp && XtIsRealized(*wp)) {
    XtDestroyWidget(*wp);
    *wp = NULL;
  }
}

/*
 * This function makes the length of two widgets equal to the widest of them
 */
extern void
make_equal_width(Widget w1, Widget w2)
{
  Dimension width1, width2;
  
  XtVaGetValues(w1, XtNwidth, &width1, NULL);
  XtVaGetValues(w2, XtNwidth, &width2, NULL);
  XtVaSetValues(w1, XtNwidth, MAX(width1, width2), NULL);
  XtVaSetValues(w2, XtNwidth, MAX(width1, width2), NULL);
}
