/* Copyright (C) 1992 Bengt Martensson and Institute for Dynamical
Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: main.cc,v 1.28 1994/11/01 12:52:38 bengt Exp bengt $

// This file contains the main routine, all X setup stuff, command line
// encoding, resources, and all the workproc stuff.

#ifdef __GNUG__
#pragma implementation "gnans.h"
#endif

#ifndef NOSTRIPCHART
#define STRIPCHART
#endif

#include <stdlib.h>
#include <unistd.h>

#include "externals.h"
#include "siteprefs.h"
#include "utillib.h"
#include "CIntegrator.h"
#include "Resource.h"
#include "PlotterControl.h"

#define String XtString
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>


#ifdef USEXAW3DINCLUDES

#include <X11/Xaw3d/Label.h>
#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Cardinals.h>
#include <X11/Xaw3d/Paned.h>
#include <X11/Xaw3d/StripChart.h>

#else

#include <X11/Xaw/Label.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/StripChart.h>

#endif

#undef String 
#include <string> 
#include <iostream>		// Probably bug in libg++2.4
#include <fstream>

#ifndef X_OK 
#define X_OK 01 
#endif

// Forward declarations 
static void create_lower_row(); 
static void syntax();
#ifdef STRIPCHART
static void how_far(Widget w, XtPointer client_data, XtPointer value);
#endif
static void enable_run_buttons(bool busy);
static void conditionally_enable_button(Widget button, bool enable);
static Boolean work_proc();
static bool test_args_for_X(int argc, char *argv[]);
static int options(int argc, char *argv[]);
static void check_create_lockfile();
static void run_startup_file();

// Global (for the file) widgets
static Widget outer;
static Widget button_quit;
static Widget button_start;
static Widget button_stop;
static Widget button_continue;
static Widget button_script; 
static Widget button_plot;
static Widget button_about;
#ifdef CHILDREN
static Widget button_children;
extern void 
Children(Widget mother_widget, XtPointer client_data, XtPointer call_data);
#endif
#ifdef HELP
static Widget button_help;
#endif
static Widget button_display;
static Widget button_parameters;
static Widget button_states;
#ifdef DYNAMIC_LOADING
static Widget button_system;
#endif

static Widget button_functions;
static Widget button_seeds;
static Widget button_settings;
static Widget button_times;

static Widget cpu;
#ifdef WORKING_MEMORY_GAUGE
static Widget bytes;
#endif
#ifdef STRIPCHART
static Widget stripchart;
#endif

// If true, suppress all prompts and greeting messages.
bool be_quiet = false;

// Number of microseconds cpu at the start of the last integration.
static double last_cpu = 0;

// Extra, not widget bound, application resources
typedef struct
{
  float starttime;
  float stoptime;
  float timestep;
  float storeincrement;
  float wienertick;
  float relerror;
  float abserror;
  XtString floating_point_format;
  XtString default_store_format;
  XtString default_algorithm;
  XtString default_stochastic_algorithm;
  float default_stochastic_timestep;
  XtString cmd_prompt;
  bool do_logging;
} AppResources;

AppResources app_resources;

static float _default_starttime_[] = { DEFAULT_STARTTIME };
static float _default_stoptime_[] = { DEFAULT_STOPTIME };
static float _default_timestep_[] = { DEFAULT_TIMESTEP };
static float _default_storeincrement_[] = { DEFAULT_STOREINCREMENT };
static float _default_relerror_[] = { DEFAULT_RELERROR };
static float _default_abserror_[] = { DEFAULT_ABSERROR };
static float _default_wienertick_[] = { DEFAULT_WIENERTICK };
static char _default_fp_format_[] = DEFAULT_FP_FORMAT;
static char _default_store_format_[] = DEFAULT_STORE_FORMAT;
static char _default_algorithm_[] = DEFAULT_ALGORITHM;
static char _default_stochastic_algorithm_[] = DEFAULT_STOCHASTIC_ALGORITHM;
static float _default_stochastic_timestep_[] = { DEFAULT_STOCHASTIC_TIMESTEP };
static char _default_cmd_prompt_[] = DEFAULT_CMD_PROMPT;
static bool _do_logging_[] = { DO_LOGGING };

static XtResource resources[] = {
  {"starttime", "Starttime", XtRFloat, sizeof(float),
     XtOffset(AppResources *, starttime), XtRFloat,
     (XtPointer) _default_starttime_},
  {"stoptime", "Stoptime", XtRFloat, sizeof(float),
     XtOffset(AppResources *, stoptime), XtRFloat,
     (XtPointer) _default_stoptime_},
  {"timestep", "Timestep", XtRFloat, sizeof(float),
     XtOffset(AppResources *, timestep), XtRFloat,
     (XtPointer) _default_timestep_},
  {"storeincrement", "Storeincrement", XtRFloat, sizeof(float),
     XtOffset(AppResources *, storeincrement), XtRFloat,
     (XtPointer) _default_storeincrement_},
  {"wienertick", "Wienertick", XtRFloat, sizeof(float),
     XtOffset(AppResources *, wienertick), XtRFloat,
     (XtPointer) _default_wienertick_},
  {"relerror", "Relerror", XtRFloat, sizeof(float),
     XtOffset(AppResources *, relerror), XtRFloat,
     (XtPointer) _default_relerror_},
  {"abserror", "Abserror", XtRFloat, sizeof(float),
     XtOffset(AppResources *, abserror), XtRFloat,
     (XtPointer) _default_abserror_},
  {"floating_point_format", "Floating_Point_Format", XtRString,
     sizeof(XtString),
     XtOffset(AppResources *, floating_point_format), XtRString,
     (XtPointer) _default_fp_format_},
  {"default_store_format", "Default_Store_Format", XtRString,
     sizeof(XtString),
     XtOffset(AppResources *, default_store_format), XtRString,
     (XtPointer) _default_store_format_},
  {"default_algorithm", "Default_Algorithm", XtRString, sizeof(XtString),
     XtOffset(AppResources *, default_algorithm), XtRString,
     (XtPointer) _default_algorithm_},
  {"default_stochastic_algorithm", "Default_Stochastic_Algorithm",
     XtRString, sizeof(XtString),
     XtOffset(AppResources *, default_stochastic_algorithm), XtRString,
     (XtPointer) _default_stochastic_algorithm_},
  {"default_stochastic_timestep", "Default_Stochastic_Timestep",
     XtRFloat, sizeof(float),
     XtOffset(AppResources *, default_stochastic_timestep), XtRFloat,
     (XtPointer) _default_stochastic_timestep_},
  {"cmd_prompt", "Cmd_Prompt", XtRString, sizeof(XtString),
     XtOffset(AppResources *, cmd_prompt), XtRString,
     (XtPointer) _default_cmd_prompt_},
  {"do_logging", "Do_Logging", XtRBoolean, sizeof(Boolean),
     XtOffset(AppResources *, do_logging), XtRBoolean,
     (XtPointer) _do_logging_},
};

XtString fallback_resources[] = { 
#include "Gnans.ad.h"
    NULL,
};

// Table of action procedures which can be called as actions
XtActionsRec actionTable[] = {
    {"change_value_action", change_value_action},
    {"fileselector_open_action", fileselector_open_action},
    {"destroy_popup_action", destroy_popup_action},
};

/****************************************************************/
// Sets up default values for internal parameters etc.  This is the X
// version, which uses resources.  If the app-defaults are not found,
// just quit.  (Just to force people to install the app-defaults in order
// to prevent stupid questions and "bugreports".)


static void
assign_defaults_x()
{
  starttime = app_resources.starttime;
  systemtime = double(starttime);
  stoptime = app_resources.stoptime;
  timestep = app_resources.timestep;
  storeincrement = app_resources.storeincrement;
  wienertick = app_resources.wienertick;
  relerror = app_resources.relerror;
  abserror = app_resources.abserror;
  default_algorithm = app_resources.default_algorithm;
  default_stochastic_algorithm = app_resources.default_stochastic_algorithm;
  default_stochastic_timestep = app_resources.default_stochastic_timestep;
  floating_point_format = app_resources.floating_point_format;
  default_store_format = app_resources.default_store_format;
  select_format(default_store_format.c_str());
  cmd_prompt = app_resources.cmd_prompt;
  if (default_algorithm == "")
    gnans_error(just_say_no,
		"Application defaults not found or not readable, quitting");
  do_logging = app_resources.do_logging;
  switches_table = switches_table_standard;
}

// Sets up default values for internal parameters etc.  This is the tty
// version, invoked if we run Gnans withour X support, e.g. via pipes.
// It takes the defaults from the compiled-in values found in
// ../config/site.h.

static void
assign_defaults_tty()
{
  starttime = DEFAULT_STARTTIME;
  systemtime = DEFAULT_STARTTIME;
  stoptime = DEFAULT_STOPTIME;
  timestep = DEFAULT_TIMESTEP;
  storeincrement = DEFAULT_STOREINCREMENT;
  wienertick = DEFAULT_WIENERTICK;
  relerror = DEFAULT_RELERROR;
  abserror = DEFAULT_ABSERROR;

  default_algorithm = DEFAULT_ALGORITHM;
  default_stochastic_algorithm = DEFAULT_STOCHASTIC_ALGORITHM;
  default_stochastic_timestep = DEFAULT_STOCHASTIC_TIMESTEP;
  floating_point_format = DEFAULT_FP_FORMAT;
  select_format(DEFAULT_STORE_FORMAT);
  cmd_prompt = DEFAULT_CMD_PROMPT;
  do_logging = DO_LOGGING;
  switches_table = switches_table_standard;
}

/****************************************************************/
// Now for some support routines

// Enable buttons for idle status (Start yes, Stop no, e.g.)
extern void
enable_buttons_idle()
{
  enable_run_buttons(false);
}

// Enable buttons for idle status (Start no, Stop yes, e.g.)
extern void
enable_buttons_busy()
{
  enable_run_buttons(true);
}

// Support routine for the two above
static void
enable_run_buttons(bool busy)
{
  conditionally_enable_button(button_start, ! busy);
  conditionally_enable_button(button_stop, busy);
  conditionally_enable_button(button_continue, ! busy);
  conditionally_enable_button(button_script, ! busy);
  conditionally_enable_button(button_plot, ! busy);
  PlotterControl::enable_plotter_controls(! busy);
#ifdef DYNAMIC_LOADING
  conditionally_enable_button(button_system, ! busy);
#endif
}

// Support routine for the one above. Enables the button (first
// argument) iff second argument is true.
static void
conditionally_enable_button(Widget button, bool enable)
{
  if (X)
    XtVaSetValues(button, XtNsensitive, (int) enable, NULL);
}

// Enables buttons (States, Parameters, Functions, Seeds) iff it
// makes sense.
extern void disable_some_buttons()
{

  conditionally_enable_button(button_states,
			      current_system.number_of_states > 0);
  conditionally_enable_button(button_parameters,
			      current_system.number_of_parameters > 0);
  conditionally_enable_button(button_functions,
			      current_system.number_of_functions > 0);
  conditionally_enable_button(button_seeds,
			      current_system.number_of_randoms > 0);
}

// Updates the title of the toplevel widget.
// Called after a system change :-)
extern void 
set_toplevel_title()
{
  std::string title = "Gnans@" + gethostname_as_string()
    + "[" + current_system.system_name + "]";
  XtVaSetValues(toplevel, XtNtitle, title.c_str(), NULL);
}

/****************************************************************/
// Now the routines which are setting up the main window.
static void
create_upper_row()
{
  Widget box;

  box = XtVaCreateManagedWidget("box", boxWidgetClass, outer, NULL);
  button_start = XtVaCreateManagedWidget("start", commandWidgetClass, box,
					 NULL);
  XtAddCallback(button_start, XtNcallback, Start, NULL);

  button_stop = XtVaCreateManagedWidget("stop", commandWidgetClass, box, NULL);
  XtAddCallback(button_stop, XtNcallback, Stop, NULL);

  button_continue = XtVaCreateManagedWidget("continue", commandWidgetClass,
					    box, NULL);
  XtAddCallback(button_continue, XtNcallback, Continue, NULL);

#ifdef DYNAMIC_LOADING
  button_system = XtCreateManagedWidget("system", commandWidgetClass,
					box, (ArgList) NULL, ZERO);
  XtAddCallback(button_system, XtNcallback, System, NULL);
#endif

  button_script = XtCreateManagedWidget("script", commandWidgetClass,
					box, (ArgList) NULL, ZERO);
  XtAddCallback(button_script, XtNcallback, Script, NULL);

  button_plot = XtVaCreateManagedWidget("plot", commandWidgetClass, box, NULL);
  XtAddCallback(button_plot, XtNcallback, Plot, NULL);
  button_display = XtVaCreateManagedWidget("display", commandWidgetClass,
					   box, NULL);
  XtAddCallback(button_display, XtNcallback, Display_vars_callback, NULL);
#ifdef CHILDREN
  button_children = XtVaCreateManagedWidget("children", commandWidgetClass,
					    box, NULL);
  XtAddCallback(button_children, XtNcallback, Children, NULL);
#endif
  button_about = XtVaCreateManagedWidget("about", commandWidgetClass,
					 box, NULL);
  XtAddCallback(button_about, XtNcallback, About, NULL);
  
#ifdef HELP
  button_help = XtVaCreateManagedWidget("help", commandWidgetClass, box, NULL);
  XtAddCallback(button_help, XtNcallback, Help, NULL);
#endif
}

static void create_middle_row()
{
  Widget box;
  box = XtVaCreateManagedWidget("box", boxWidgetClass, outer, NULL);
  button_states = XtVaCreateManagedWidget("states", commandWidgetClass,
					  box, NULL);
  XtAddCallback(button_states, XtNcallback, States, NULL);

  button_parameters = XtVaCreateManagedWidget("parameters",
					      commandWidgetClass, box, NULL);
  XtAddCallback(button_parameters, XtNcallback, Parameters, NULL);

  button_functions = XtVaCreateManagedWidget("functions", commandWidgetClass,
					     box, NULL);
  XtAddCallback(button_functions, XtNcallback, Functions, NULL);
  button_seeds = XtVaCreateManagedWidget("seeds", commandWidgetClass,
					 box, NULL);
  XtAddCallback(button_seeds, XtNcallback, Seeds, NULL);

  button_times = XtVaCreateManagedWidget("times", commandWidgetClass,
					 box, NULL);
  XtAddCallback(button_times, XtNcallback, Times, NULL);
  button_settings = XtVaCreateManagedWidget("settings", commandWidgetClass,
					    box, NULL);
  XtAddCallback(button_settings, XtNcallback, Settings, NULL);
}

static void
create_quit_row()
{
  Widget box;
  box = XtVaCreateManagedWidget("box", boxWidgetClass, outer, NULL);
  button_quit = XtVaCreateManagedWidget("quit", commandWidgetClass, box, NULL);
  XtAddCallback(button_quit, XtNcallback, Quit, NULL);
}

static void
create_lower_row()
{
  Widget box;
  box = XtVaCreateManagedWidget("box", boxWidgetClass, outer, NULL);
  cpu = XtVaCreateManagedWidget("cpu", labelWidgetClass, box, NULL);
#ifdef WORKING_MEMORY_GAUGE
  bytes = XtVaCreateManagedWidget("bytes", labelWidgetClass, box, NULL);
#endif
  indicate_progress();
#ifdef STRIPCHART
  stripchart = XtVaCreateManagedWidget("stripchart", stripChartWidgetClass,
				       box, NULL);
  XtAddCallback(stripchart, XtNgetValue, how_far, NULL);
#if 0
  static Cursor wascursor;
  XtVaGetValues(stripchart, XtNcursorName, &wascursor, NULL);
  gnans_error(debug, "`%s'", wascursor);
  XtVaSetValues(stripchart, XtNcursorName, "shuttle", NULL);
  //XtVaGetValues(stripchart, XtNcursorName, &wascursor, NULL);
 //gnans_error(debug, "`%s'", wascursor);
 Arg arg[1];
  XtSetArg(arg[0], XtNcursorName, NULL);
  XtSetValues(stripchart, arg, ONE);
#endif
#endif
}

// The following code is more or less verbatim from the X FAQ, question 132.
static Window busy_window = (Window) NULL;

static void
create_busy_window()
{
  unsigned long valuemask;
  XSetWindowAttributes attributes;

  /* Ignore device events while the busy cursor is displayed. */
  valuemask = CWDontPropagate | CWCursor;
  attributes.do_not_propagate_mask =  (KeyPressMask | KeyReleaseMask
				       | ButtonPressMask | ButtonReleaseMask
				       | PointerMotionMask);
  attributes.cursor = XCreateFontCursor(XtDisplay(toplevel), XC_watch);

  /* The window will be as big as the display screen, and clipped by
     its own parent window, so we never have to worry about resizing */
  busy_window = XCreateWindow(XtDisplay(toplevel), XtWindow(toplevel), 0, 0,
			      65535, 65535, (unsigned int) 0, 0, InputOnly,
			      CopyFromParent, valuemask, &attributes);
}

void
enable_busy_window()
{
  if (! X)
    return;
  if (! busy_window)
    create_busy_window();
  XMapRaised(XtDisplay(toplevel), busy_window);
  XFlush(XtDisplay(toplevel));
}

void
disable_busy_window()
{
  if (! X)
    return;
  XUnmapWindow(XtDisplay(toplevel), busy_window);
  XFlush(XtDisplay(toplevel));
}

/**************** Progress indicator and friends ****************/
// This is the main routine in the progress indicator. It is to be called
// regularly.  It does not do what I want with respect to the memory
// consumption.
extern void
indicate_progress()
{
  if (!X)
    return;
  
  Resource rusage;		// Constructor calls getrusage or equiv.
  Dimension width = 0;
  std::string tim = number_String(int(rusage.user_seconds() + rusage.system_seconds()
			- last_cpu));
  std::string cpu_string = "Cpu: " + tim;
  tim = number_String(int(rusage.user_seconds() + rusage.system_seconds()));
  cpu_string += " tot: " + tim;
  XtVaGetValues(button_start, XtNwidth, &width, NULL);
  XtVaSetValues(cpu, XtNlabel, cpu_string.c_str(), XtNwidth, width, NULL);
#ifdef WORKING_MEMORY_GAUGE
  std::string byt = number_String(rusage.kilobytes());
  std::string bytes_string = "Mem: " + byt + " kB";
  XtVaSetValues(bytes, XtNlabel, bytes_string.c_str(), XtNwidth, width,
		NULL);
#endif
}

// "Zeros" cpu used in current run.
extern void
reset_cpu()
{
#if 0
  static struct rusage this_run;
  if (getrusage(RUSAGE_SELF, &this_run))
    gnans_error(impossible, "Error by getrusage");
  last_cpu = this_run.ru_utime.tv_sec + this_run.ru_stime.tv_sec;
#else
  Resource this_run;
  last_cpu = this_run.user_seconds() + this_run.system_seconds();
#endif
}

static inline double
min(double a, double b)
{
  return a < b ? a : b;
}

#ifdef STRIPCHART
// This function is called by the stripchart widget.  Returns a
// number between 0 and 10, indicating how far the run has proceeded.

const double MULTIPLIER = 9.999;
static void
how_far(Widget /*w*/, XtPointer /*client_data*/, XtPointer value)
{
  *((double *) value) =
    workprocs
    ? MULTIPLIER * min((systemtime - starttime)/(stoptime - starttime), 1.0)
    : MULTIPLIER;
}
#endif

/**************** Workproc stuff ****************/ 
// Some global variables
static XtAppContext app_con;
static XtWorkProcId Proc_ID;

// Pointer to function doing our work (e.g. integrate_lsode) 
static bool (*work_loop)(bool initialize);

// Is this a virgin integration?  
// In the call of the work_loop, we set continue = false if this is true
static bool virgin = true;

// This function is called to register the function (second argument)
// as a procedure being called by the workproc. 
// First argument is continuation in the sense of integrate.cc.

extern void
register_work_proc(bool continuation,
		   bool (*proc)(bool initialize)/* = NULL */)
{
  set_alarm();
  virgin = ! continuation;
  has_been_alarmed = false;
  work_loop = proc;
  Proc_ID = XtAppAddWorkProc(app_con, (XtWorkProc) &work_proc, NULL);
}

// This is the true workproc. It calls our work_loop, which returns
// either with true, in which case the task is finished, or with false,
// in which case the work_loop was interrupted by an alarm.  In the
// latter case, return FALSE, which causes the function to be called
// again.

static Boolean
work_proc()
{
  bool finished;
  has_been_alarmed = false;
  finished = work_loop
    ? (*work_loop)(virgin)
    : current_algorithm->loop(virgin);
  virgin = false;
  if (finished) {
    cancel_alarm();
    indicate_progress();
    if (trace_workproc)
      gnans_error(debug, "Workproc: %g\t(finished)", double(systemtime));
    if (! running_script) integrate_finish();
    running_integration = false;
    running_script = false;
    enable_buttons_idle();
    return TRUE;
  } else {
    indicate_progress();
    if (trace_workproc)
      gnans_error(debug, "Workproc: %g\t(not finished)", double(systemtime));
    return FALSE;
  }
}

/****************************************************************/
// This is the main routine. It works pretty differently if using X or not.
extern int
main(int argc, char *argv[])
{
#ifdef MALLOC_DEBUG
  union dbmalloptarg m;
  m.i = 0;
  dbmallopt(MALLOC_ZERO, &m);
  m.i = 131;
  dbmallopt(MALLOC_WARN, &m);
  m.str = "MALLOC.LOG";
  dbmallopt(MALLOC_ERRFILE, &m);
#endif
  int number_of_options;

  progname = basename(argv[0]);
  scriptfilename = "";
#ifdef DYNAMIC_LOADING
  system_filename = "";
#endif

  // Check write permissions for cwd
  char cwd[MAXSTRINGSIZE];
  getcwd(cwd, MAXSTRINGSIZE);
  if (!cwd)
    gnans_error(just_say_no, "Problems getting current working directory.");
  if (access(cwd, W_OK) != 0)
    gnans_error(just_say_no,
		"No write permissions for the current directory!\n"
		"Please cd to a writeable directory.");

  X = test_args_for_X(argc, argv);
  if (X) {
    // X interface*/
    toplevel = XtAppInitialize(&app_con, "Gnans", NULL, ZERO,
#if XtSpecificationRelease == 4
			       (Cardinal *)
#endif
			       &argc, argv, fallback_resources, NULL, ZERO);
    XtGetApplicationResources(toplevel, (XtPointer) &app_resources,
			      resources, XtNumber(resources), NULL, ZERO );
    XtAppAddActions(app_con, actionTable, XtNumber(actionTable));
    assign_defaults_x();
    number_of_options = options(argc, argv);
    check_create_lockfile();
    running_script = false;

    debug_setup();
    signal_setup();
    integrators_setup();
    
    outer = XtVaCreateManagedWidget("paned", panedWidgetClass, toplevel, NULL);
    // Action buttons 
    create_upper_row();

    // Settings
    create_middle_row();

    // The Quit button
    create_quit_row();

    // Monitoring
    create_lower_row();

    setup_system_static();

    if (do_logging)
      open_log();

#ifdef DYNAMIC_LOADING
    if (number_of_options == argc)
      ;
    else if (number_of_options == argc - 1) {
      system_filename = argv[number_of_options];
      setup_system_dynamic(std::string(system_filename));
    } else
      syntax();
#else				// ! DYNAMIC_LOADING
    if (number_of_options != argc)
      syntax();
#endif				// ! DYNAMIC_LOADING
    XtRealizeWidget(toplevel);
    enable_buttons_idle();

    // Don't use workprocs yet 
    bool workprocs_saved = workprocs;
    workprocs = false;
    run_startup_file();
    if (scriptfilename != "")
      decode_commands(scriptfilename.c_str());
    workprocs = workprocs_saved;
    
    XtAppMainLoop(app_con);
  } else {
    // tty interface
    assign_defaults_tty();
    number_of_options = options(argc, argv);
    check_create_lockfile();
    debug_setup();
    signal_setup();
    integrators_setup();
    setup_system_static();
    if (do_logging)
      open_log();

#ifdef DYNAMIC_LOADING
    if (number_of_options == argc)
      ;
    else if (number_of_options == argc - 1)
      setup_system_dynamic(std::string(argv[number_of_options]));
    else
      syntax();
#else // ! DYNAMIC_LOADING
    if (number_of_options != argc)
      syntax();
#endif // ! DYNAMIC_LOADING

    setbuf(stdout, NULL);
    run_startup_file();
    if (scriptfilename != "")
      decode_commands(scriptfilename.c_str());
    if (! be_quiet)
      About(NULL, NULL, NULL);
    decode_commands("");
  }
  die("", 0);
}
/****************************************************************/
// Check if there is a readable startup file, and execute it if so.
static void
run_startup_file()
{
  std::string filename = std::string(getenv("HOME")) + "/" + GNANSSTARTUPFILENAME;
  bool success = access(filename.c_str(), R_OK) == 0;
  if (do_logging)
    va_log_event(success
		 ? "Found startup file ``%s''"
		 : "No (readable) file named ``%s'' found.",
		 filename.c_str());
  if (success)
    decode_commands(filename.c_str());
}

static void
check_create_lockfile()
{
  std::ifstream oldlockfile(GNANSLOCKFILENAME);
  if (!oldlockfile) {
    std::ofstream lockfile(GNANSLOCKFILENAME);
    char cwd[MAXSTRINGSIZE];
    getcwd(cwd, MAXSTRINGSIZE); 
    lockfile << "Gnans was started in this directory by " << getenv("USER")
	     << "@" << getenv("HOST") << "\nat " << getdate()
	     << " (pid " << getpid() << ").\n";
    lockfile << "If that process does not exist any longer, "
      "please remove the file " << cwd << "/"
				<< GNANSLOCKFILENAME << ".\n";
    // Closed by destructor
  } else {
    std::string mess("Only one Gnans can run in one single directory at a time.\n");
    std::string word;
    while (oldlockfile >> word)
      mess += word + " ";
    gnans_error(lock_exists, mess.c_str());
  }
}

// Tries to find out if X interface is explicitly requested,
// or explicitly not requested. If neither X nor tty is requested,
// find out if I can open a display or not.
static bool
test_args_for_X(int argc, char *argv[])
{
  int i;
  Display *dpy;
  bool X = false;
  bool tty = false;
  std::string the_display = "";

  tty = progname.find("tty") != std::string::npos;

  for (i = 1; i < argc; i++) {
    if (std::string(argv[i]) == "-display") {
      if (i == argc - 1)
	syntax();
      the_display = argv[i+1];
      X = true;
    }
    tty = tty || (std::string(argv[i]) == "-nw");
  }
  
  if (tty)
    if (X) {			// User wants both tty and X
      gnans_error(just_say_no, "You cannot use BOTH X AND tty interface");
      return false;		// never reached
    } else
      return false;
  else {			// ! tty
    dpy = XOpenDisplay((char*)(!(the_display.empty()) ? the_display.c_str() : NULL));
    if (dpy) {
      XCloseDisplay(dpy);
      return true;
    }
    else {
      gnans_error(message, "Could not open display, using tty interface");
      return false;
    }
  }
}

// Command line decoding. Based on (a possibly obsolete version of)
// Gnu getopt, using long option names. 
#include "getopt.h"

#define NOARG  0
#define ARG    1
#define OPTARG 2

static struct option long_options[] = {
  {"debug",	ARG,	0, 1},
  {"version", 	NOARG,  0, 2},
  {"nw", 	NOARG,  0, 3},
  {"scriptfile", ARG,    0, 4},
  {"development", NOARG, 0, 5},
  {"expertmode", NOARG,  0, 6},
  {"help",	NOARG,  0, 7},
  {"?",    	NOARG,  0, 8},
  {"quiet",    	NOARG,  0, 9},
  
  {0, 0, 0, 0}
};

// Do the actual decoding. Returns the number of arguments read.
static int 
options(int argc, char *argv[])
{
  int c, option_index = -1;

  // Parse options
  while ((c = getopt_long_only(argc, argv, "", long_options, &option_index))
	 != EOF) {
    switch (c) {
    case 1:
      gnansdebug = atoi(optarg);
      break;
    case 2:
      gnans_error(just_say_no, "Gnans version %s, compiled %s",
		  version_string, compiled_string);
      break;
    case 3:
      break;
    case 4:
      scriptfilename = std::string(optarg);
      break;
    case 5:
      development = true;
      update_development_status();
      break;
    case 6:
      expert_mode = true;
      break;
    case 9:
      be_quiet = true;
      cmd_prompt = "";
      floating_point_format = " %.16e\n";	// This is a kludge.
      break;
    default:
      syntax();
      break;
    }
  }
  return optind;
}

// Called on erroneous calling syntax. Just barf.
static void 
syntax()
{
  gnans_error(just_say_no,
	      "Usage: %s [-Xtoptions] [-options] [system]\n",
	      progname.c_str());
}
